@@ -5,7 +5,7 @@ gmr
 Gaussian Mixture Models (GMMs) for clustering and regression in Python.
 """
 
-__version__ = "1.5.1"
+__version__ = "1.6"
 
 try:
     # Idea from sklearn:
@@ -20,7 +20,7 @@ except NameError:
 if not __GMR_SETUP__:
     from . import gmm, mvn, utils
 
-    __all__ = ["gmm", "mvn", "utils"]
+    __all__ = ["gmm", "mvn", "utils", "sklearn"]
 
     from .mvn import MVN, plot_error_ellipse
     from .gmm import (GMM, plot_error_ellipses, kmeansplusplus_initialization,
