@@ -1,7 +1,9 @@
 import numpy as np
-from .utils import check_random_state, pinvh
+from .utils import check_random_state
 import scipy as sp
 from scipy.stats import chi2
+from scipy.spatial.distance import mahalanobis
+from scipy.linalg import pinvh
 
 
 def invert_indices(n_features, indices):
@@ -20,10 +22,10 @@ class MVN(object):
 
     Parameters
     ----------
-    mean : array, shape (n_features), optional
+    mean : array-like, shape (n_features), optional
         Mean of the MVN.
 
-    covariance : array, shape (n_features, n_features), optional
+    covariance : array-like, shape (n_features, n_features), optional
         Covariance of the MVN.
 
     verbose : int, optional (default: 0)
@@ -41,6 +43,11 @@ class MVN(object):
         self.random_state = check_random_state(random_state)
         self.norm = None
 
+        if self.mean is not None:
+            self.mean = np.asarray(self.mean)
+        if self.covariance is not None:
+            self.covariance = np.asarray(self.covariance)
+
     def _check_initialized(self):
         if self.mean is None:
             raise ValueError("Mean has not been initialized")
@@ -55,6 +62,9 @@ class MVN(object):
         X : array-like, shape (n_samples, n_features)
             Samples from the true function.
 
+        bessels_correction : bool
+            Apply Bessel's correction to the covariance estimate.
+
         Returns
         -------
         self : MVN
@@ -133,8 +143,11 @@ class MVN(object):
         n_dof = len(x) - 1
         return self.squared_mahalanobis_distance(x) <= chi2(n_dof).ppf(alpha)
 
-    def to_probability_density(self, X):
-        """Compute probability density.
+    def to_norm_factor_and_exponents(self, X):
+        """Compute normalization factor and exponents of Gaussian.
+
+        These values can be used to compute the probability density function
+        of this Gaussian: p(x) = norm_factor * np.exp(exponents).
 
         Parameters
         ----------
@@ -143,8 +156,12 @@ class MVN(object):
 
         Returns
         -------
-        p : array, shape (n_samples,)
-            Probability densities of data.
+        norm_factor : float
+            Normalization factor: constant term outside of exponential
+            function in probability density function of this Gaussian.
+
+        exponents : array, shape (n_samples,)
+            Exponents to compute probability density function.
         """
         self._check_initialized()
 
@@ -173,7 +190,25 @@ class MVN(object):
         X_normalized = sp.linalg.solve_triangular(
             L, X_minus_mean.T, lower=True).T
 
-        return self.norm * np.exp(-0.5 * np.sum(X_normalized ** 2, axis=1))
+        exponent = -0.5 * np.sum(X_normalized ** 2, axis=1)
+
+        return self.norm, exponent
+
+    def to_probability_density(self, X):
+        """Compute probability density.
+
+        Parameters
+        ----------
+        X : array-like, shape (n_samples, n_features)
+            Data.
+
+        Returns
+        -------
+        p : array, shape (n_samples,)
+            Probability densities of data.
+        """
+        norm_factor, exponents = self.to_norm_factor_and_exponents(X)
+        return norm_factor * np.exp(exponents)
 
     def marginalize(self, indices):
         """Marginalize over everything except the given indices.
@@ -209,7 +244,8 @@ class MVN(object):
             Conditional MVN distribution p(Y | X=x).
         """
         self._check_initialized()
-        mean, covariance = self._condition(
+        mean, covariance = condition(
+            self.mean, self.covariance,
             invert_indices(self.mean.shape[0], indices), indices, x)
         return MVN(mean=mean, covariance=covariance,
                    random_state=self.random_state)
@@ -221,10 +257,10 @@ class MVN(object):
 
         Parameters
         ----------
-        indices : array, shape (n_features_1,)
+        indices : array-like, shape (n_features_1,)
             Indices of dimensions that we want to condition.
 
-        X : array, shape (n_samples, n_features_1)
+        X : array-like, shape (n_samples, n_features_1)
             Values of the features that we know.
 
         Returns
@@ -236,19 +272,11 @@ class MVN(object):
             Covariance of the predicted features.
         """
         self._check_initialized()
-        return self._condition(invert_indices(self.mean.shape[0], indices),
-                               indices, X)
-
-    def _condition(self, i1, i2, X):
-        cov_12 = self.covariance[np.ix_(i1, i2)]
-        cov_11 = self.covariance[np.ix_(i1, i1)]
-        cov_22 = self.covariance[np.ix_(i2, i2)]
-        prec_22 = pinvh(cov_22)
-        regression_coeffs = cov_12.dot(prec_22)
-
-        mean = self.mean[i1] + regression_coeffs.dot((X - self.mean[i2]).T).T
-        covariance = cov_11 - regression_coeffs.dot(cov_12.T)
-        return mean, covariance
+        indices = np.asarray(indices, dtype=int)
+        X = np.asarray(X)
+        return condition(
+            self.mean, self.covariance,
+            invert_indices(self.mean.shape[0], indices), indices, X)
 
     def squared_mahalanobis_distance(self, x):
         """Squared Mahalanobis distance between point and this MVN.
@@ -262,8 +290,8 @@ class MVN(object):
         d : float
             Squared Mahalanobis distance
         """
-        d = x - self.mean
-        return d.dot(np.linalg.inv(self.covariance)).dot(d)
+        self._check_initialized()
+        return mahalanobis(x, self.mean, np.linalg.inv(self.covariance)) ** 2
 
     def to_ellipse(self, factor=1.0):
         """Compute error ellipse.
@@ -443,3 +471,73 @@ def plot_error_ellipse(ax, mvn, color=None, alpha=0.25,
         if color is not None:
             ell.set_color(color)
         ax.add_artist(ell)
+
+
+def regression_coefficients(covariance, i1, i2, cov_12=None):
+    """Compute regression coefficients to predict conditional distribution.
+
+    Parameters
+    ----------
+    covariance : array, shape (n_features, n_features)
+        Covariance of MVN
+
+    i1 : array, shape (n_features1,)
+        Input feature indices
+
+    i2 : array, shape (n_features2,)
+        Output feature indices
+
+    cov_12 : array, shape (n_features1, n_features2), optional (default: None)
+        Precomputed block of the covariance matrix between input features and
+        output features
+
+    Returns
+    -------
+    regression_coeffs : array, shape (n_features1, n_features2)
+        Regression coefficients. These can be used to compute the mean of the
+        conditional distribution as
+        mean[i1] + regression_coeffs.dot((X - mean[i2]).T).T
+    """
+    if cov_12 is None:
+        cov_12 = covariance[np.ix_(i1, i2)]
+    cov_22 = covariance[np.ix_(i2, i2)]
+    prec_22 = pinvh(cov_22)
+    return cov_12.dot(prec_22)
+
+
+def condition(mean, covariance, i1, i2, X):
+    """Compute conditional mean and covariance.
+
+    Parameters
+    ----------
+    mean : array, shape (n_features,)
+        Mean of MVN
+
+    covariance : array, shape (n_features, n_features)
+        Covariance of MVN
+
+    i1 : array, shape (n_features1,)
+        Input feature indices
+
+    i2 : array, shape (n_features2,)
+        Output feature indices
+
+    X : array, shape (n_samples, n_features1)
+        Inputs
+
+    Returns
+    -------
+    mean : array, shape (n_features2,)
+        Mean of the conditional distribution
+
+    covariance : array, shape (n_features2, n_features2)
+        Covariance of the conditional distribution
+    """
+    cov_12 = covariance[np.ix_(i1, i2)]
+    cov_11 = covariance[np.ix_(i1, i1)]
+    regression_coeffs = regression_coefficients(
+        covariance, i1, i2, cov_12=cov_12)
+
+    mean = mean[i1] + regression_coeffs.dot((X - mean[i2]).T).T
+    covariance = cov_11 - regression_coeffs.dot(cov_12.T)
+    return mean, covariance
